from unittest.mock import call, patch, seal

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.process.models import ProcessMode
from arkindex.project.tests import FixtureAPITestCase


@override_settings(PONOS_MAXIMUM_TASK_TTL=3600)
class TestProcessRetrieve(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(
            mode=ProcessMode.Files,
            creator=cls.user,
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    @patch("arkindex.project.mixins.get_max_level")
    def test_requires_guest(self, get_max_level_mock):
        get_max_level_mock.return_value = None
        seal(get_max_level_mock)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "Not found."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_retrieve(self):
        self.client.force_login(self.user)
        # Have the endpoint return a task too
        self.process.run()
        task = self.process.tasks.get()

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "activity_state": "disabled",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "element": None,
            "element_name_contains": None,
            "element_type": None,
            "farm": None,
            "files": [],
            "folder_type": None,
            "id": str(self.process.id),
            "load_children": False,
            "ml_class_id": None,
            "mode": "files",
            "name": None,
            "state": "unscheduled",
            "tasks": [
                {
                    "id": str(task.id),
                    "depth": 0,
                    "parents": [],
                    "run": 0,
                    "shm_size": None,
                    "slug": "import_files",
                    "state": "unscheduled",
                    "requires_gpu": False,
                    "original_task_id": None,
                    "started": None,
                    "finished": None,
                    "ttl": 3600,
                }
            ],
            "template_id": None,
            "use_cache": False,
            "created": self.process.created.isoformat().replace("+00:00", "Z"),
            "updated": self.process.updated.isoformat().replace("+00:00", "Z"),
            "started": self.process.started.isoformat().replace("+00:00", "Z"),
            "finished": None,
        })
