from unittest.mock import MagicMock, patch
from uuid import uuid4

from django.core import mail
from django.urls import reverse
from rest_framework import status
from rq.job import Job

from arkindex.process.models import ActivityState, Process, ProcessMode, WorkerActivityState, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCreateProcessFailures(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            name="Borked process",
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_requires_corpus_admin(self, has_access_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to the corpus of this process."})

    def test_wrong_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": str(uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_local(self):
        """
        A local process with no corpus is treated as a process with no admin access
        """
        self.client.force_login(self.user)
        local_process = self.user.processes.get(mode=ProcessMode.Local)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": local_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to the corpus of this process."})

    def test_invalid_mode(self):
        self.client.force_login(self.user)
        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.finished = self.process.started = self.process.updated
                self.process.save()

                with self.assertNumQueries(3):
                    response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {"__all__": [
                    "The process must be in mode Workers to perform this action",
                    "The process has no element with a WorkerActivity in an Error state",
                ]})

    def test_invalid_state(self):
        self.client.force_login(self.user)
        self.process.activities.create(
            element=self.corpus.elements.get(name="Volume 1, page 1r"),
            worker_version=WorkerVersion.objects.first(),
            state=WorkerActivityState.Error,
        )

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"__all__": ["The process must be finished to perform this action"]})

    def test_no_failures(self):
        self.client.force_login(self.user)
        self.process.finished = self.process.started = self.process.updated
        self.process.save()

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"__all__": [
            "The process has no element with a WorkerActivity in an Error state"
        ]})

    @patch("arkindex.project.triggers.process_tasks.create_process_failures.delay")
    @patch("arkindex.process.tasks.get_current_job")
    def test_create(self, current_job_mock, failures_task_mock):
        from arkindex.process.tasks import create_process_failures
        failures_task_mock.side_effect = lambda process, user_id, description: create_process_failures(process)
        current_job_mock.return_value = MagicMock(spec=Job, user_id=self.user.id)

        self.client.force_login(self.user)
        self.process.finished = self.process.started = self.process.updated
        self.process.activity_state = ActivityState.Ready
        self.process.save()

        # Create worker activities, with 1 error on the volume element
        worker_version_1, worker_version_2 = WorkerVersion.objects.all()[:2]
        self.process.activities.create(
            element=self.corpus.elements.get(name="Volume 1, page 1r"),
            worker_version=worker_version_1,
            state=WorkerActivityState.Processed,
        )
        volume = self.corpus.elements.get(name="Volume 1")
        self.process.activities.create(
            element=volume, worker_version=worker_version_1, state=WorkerActivityState.Processed
        )
        self.process.activities.create(
            element=volume, worker_version=worker_version_2, state=WorkerActivityState.Error
        )

        with self.assertNumQueries(11):
            response = self.client.post(reverse("api:process-failures", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(response.content, b"")
        new_process = Process.objects.order_by("created").last()
        self.assertEqual(new_process.name, f"Failures from {self.process.id}")
        self.assertQuerySetEqual(
            new_process.elements.all(),
            [volume],
        )
        # Ensure the user received an email once async creation is done
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, [self.user.email])
        self.assertEqual(mail.outbox[0].subject, "Process created from previous failures (Borked process)")
        self.assertIn(reverse("frontend-configure-process", kwargs={"pk": new_process.id}), mail.outbox[0].body)
