
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import FeatureUsage, Worker, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestFeatureWorkerVersionRetrieve(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.init_version = WorkerVersion.objects.get(worker__slug="initialisation")

    def test_retrieve_requires_logged_in(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:feature-worker-version", kwargs={"feature": "init_elements"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        data = response.json()
        self.assertDictEqual(data, {"detail": "Not found."})

    @patch("arkindex.users.utils.filter_rights", return_value=Worker.objects.none())
    def test_retrieve_not_public(self, filter_rights_mock):
        """
        If the worker is not public, it must be executable by the user
        """
        # Make the elements initialisation worker not public
        self.init_version.worker.public = False
        self.init_version.worker.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:feature-worker-version", kwargs={"feature": "init_elements"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        data = response.json()
        self.assertDictEqual(data, {"detail": "Not found."})
        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Worker, Role.Contributor.value))

    def test_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:feature-worker-version", kwargs={"feature": "init_elements"}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "id": str(self.init_version.id),
            "configuration": {"docker": {"command": "worker-init-elements"}},
            "docker_image_iid": self.init_version.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "revision_url": None,
            "version": 1,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.init_version.worker.id),
                "name": "Elements Initialisation Worker",
                "type": "init",
                "slug": "initialisation",
                "description": "",
                "archived": False,
                "repository_url": None,
                "cost_1k_elements": "0.000",
                "cost_cpu_hour": "0.000",
                "cost_gpu_hour": "0.000",
            },
            "created": "2020-02-02T01:23:45.678000Z",
        })

    def test_no_feature_version(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:feature-worker-version", kwargs={"feature": "s3_ingest"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
            self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_not_a_feature(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:feature-worker-version", kwargs={"feature": "make_coffee"}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), {"feature": "make_coffee is not a valid Arkindex feature."})
